@@ -38,8 +38,6 @@
 /*                                                                       */
 /*************************************************************************/
 
-#include <stdio.h>
-#include <string.h>
 #include "cst_hrg.h"
 #include "cst_cart.h"
 #include "cst_tokenstream.h"
@@ -49,6 +47,8 @@
 #include "cst_synth.h"
 #include "cst_phoneset.h"
 
+CST_VAL_REGISTER_FUNCPTR(breakfunc,cst_breakfunc)
+
 #ifndef SYNTH_MODULES_DEBUG
 #define SYNTH_MODULES_DEBUG 0
 #endif
@@ -61,10 +61,10 @@
 
 static cst_utterance *tokentosegs(cst_utterance *u);
 
-const cst_synth_module synth_method_text[] = {
+static const cst_synth_module synth_method_text[] = {
     { "tokenizer_func", default_tokenization },
     { "textanalysis_func", default_textanalysis },
-    { "pos_tagger_func", NULL },
+    { "pos_tagger_func", default_pos_tagger },
     { "phrasing_func", default_phrasing },
     { "lexical_insertion_func", default_lexical_insertion },
     { "pause_insertion_func", default_pause_insertion },
@@ -73,12 +73,23 @@ const cst_synth_module synth_method_text[] = {
     { "duration_model_func", cart_duration },
     { "f0_model_func", NULL },
     { "wave_synth_func", NULL },
+    { "post_synth_hook_func", NULL },
     { NULL, NULL }
 };
 
-const cst_synth_module synth_method_tokens[] = {
+static const cst_synth_module synth_method_text2segs[] = {
+    { "tokenizer_func", default_tokenization },
     { "textanalysis_func", default_textanalysis },
-    { "pos_tagger_func", NULL },
+    { "pos_tagger_func", default_pos_tagger },
+    { "phrasing_func", default_phrasing },
+    { "lexical_insertion_func", default_lexical_insertion },
+    { "pause_insertion_func", default_pause_insertion },
+    { NULL, NULL }
+};
+
+static const cst_synth_module synth_method_tokens[] = {
+    { "textanalysis_func", default_textanalysis },
+    { "pos_tagger_func", default_pos_tagger },
     { "phrasing_func", default_phrasing },
     { "lexical_insertion_func", default_lexical_insertion },
     { "pause_insertion_func", default_pause_insertion },
@@ -87,15 +98,19 @@ const cst_synth_module synth_method_tokens[] = {
     { "duration_model_func", cart_duration },
     { "f0_model_func", NULL },
     { "wave_synth_func", NULL },
+    { "post_synth_hook_func", NULL },
     { NULL, NULL }
 };
-const cst_synth_module synth_method_phones[] = {
+
+static const cst_synth_module synth_method_phones[] = {
     { "tokenizer_func", default_tokenization },
     { "textanalysis_func", tokentosegs },
+    { "pos_tagger_func", default_pos_tagger },
     { "intonation_func", NULL },
     { "duration_model_func", cart_duration },
     { "f0_model_func", flat_prosody },
     { "wave_synth_func", NULL },
+    { "post_synth_hook_func", NULL },
     { NULL, NULL }
 };
 
@@ -145,6 +160,11 @@ cst_utterance *utt_synth_tokens(cst_utterance *u)
     return apply_synth_method(u, synth_method_tokens);
 }
 
+cst_utterance *utt_synth_text2segs(cst_utterance *u)
+{
+    return apply_synth_method(u, synth_method_text2segs);
+}
+
 cst_utterance *utt_synth_phones(cst_utterance *u)
 {
     return apply_synth_method(u, synth_method_phones);
@@ -159,23 +179,16 @@ cst_utterance *default_tokenization(cst_utterance *u)
 
     text = utt_input_text(u);
     r = utt_relation_create(u,"Token");
-    fd = ts_open_string(text);
-    fd->whitespacesymbols = 
-	get_param_string(u->features,"text_whitespace",fd->whitespacesymbols);
-    fd->singlecharsymbols = 
-	get_param_string(u->features,"text_singlecharsymbols",
-			 fd->singlecharsymbols);
-    fd->prepunctuationsymbols = 
-	get_param_string(u->features,"text_prepunctuation",
-			 fd->prepunctuationsymbols);
-    fd->postpunctuationsymbols = 
-	get_param_string(u->features,"text_pospunctuation",
-			 fd->postpunctuationsymbols);
+    fd = ts_open_string(text,
+	get_param_string(u->features,"text_whitespace",NULL),
+	get_param_string(u->features,"text_singlecharsymbols",NULL),
+	get_param_string(u->features,"text_prepunctuation",NULL),
+        get_param_string(u->features,"text_postpunctuation",NULL));
     
     while(!ts_eof(fd))
     {
 	token = ts_get(fd);
-	if (strlen(token) > 0)
+	if (cst_strlen(token) > 0)
 	{
 	    t = relation_append(r,NULL);
 	    item_set_string(t,"name",token);
@@ -194,7 +207,7 @@ cst_utterance *default_tokenization(cst_utterance *u)
 
 cst_val *default_tokentowords(cst_item *i)
 {
-	return cons_val(string_val(item_feat_string(i,"name")), NULL);
+    return cons_val(string_val(item_feat_string(i,"name")), NULL);
 }
 
 cst_utterance *default_textanalysis(cst_utterance *u)
@@ -237,7 +250,7 @@ cst_utterance *default_textanalysis(cst_utterance *u)
 cst_utterance *default_phrasing(cst_utterance *u)
 {
     cst_relation *r;
-    cst_item *w, *p;
+    cst_item *w, *p, *lp=NULL;
     const cst_val *v;
     cst_cart *phrasing_cart;
 
@@ -249,13 +262,17 @@ cst_utterance *default_phrasing(cst_utterance *u)
 	if (p == NULL)
 	{
 	    p = relation_append(r,NULL);
-	    item_set_string(p,"name","BB");
+            lp = p;
+            item_set_string(p,"name","B");
 	}
 	item_add_daughter(p,w);
 	v = cart_interpret(w,phrasing_cart);
 	if (cst_streq(val_string(v),"BB"))
 	    p = NULL;
     }
+
+    if (lp && item_prev(lp)) /* follow festival */
+        item_set_string(lp,"name","BB");
     
     return u;
 }
@@ -300,6 +317,9 @@ cst_utterance *cart_intonation(cst_utterance *u)
     cst_item *s;
     const cst_val *v;
 
+    if (feat_present(u->features,"no_intonation_accent_model"))
+        return u;  /* not all languages have intonation models */
+
     accents = val_cart(feat_val(u->features,"int_cart_accents"));
     tones = val_cart(feat_val(u->features,"int_cart_tones"));
     
@@ -319,8 +339,8 @@ cst_utterance *cart_intonation(cst_utterance *u)
 		   ffeature_string(s,"ssyl_out"),
 		   ffeature_string(s,"accent"),
 		   ffeature_string(s,"endtone")));
-
     }
+
     return u;
 }
 
@@ -331,7 +351,7 @@ const dur_stat *phone_dur_stat(const dur_stats *ds,const char *ph)
     int i;
     for (i=0; ds[i]; i++)
 	if (cst_streq(ph,ds[i]->phone))
-	return ds[i];
+            return ds[i];
 
     return ds[0];
 }
@@ -347,6 +367,9 @@ cst_utterance *cart_duration(cst_utterance *u)
 
     end = 0;
 
+    if (feat_present(u->features,"no_segment_duration_model"))
+        return u;  /* not all methods need segment durations */
+
     dur_tree = val_cart(feat_val(u->features,"dur_cart"));
     dur_stretch = get_param_float(u->features,"duration_stretch", 1.0);
     ds = val_dur_stats(feat_val(u->features,"dur_stats"));
@@ -376,30 +399,60 @@ cst_utterance *cart_duration(cst_utterance *u)
     return u;
 }
 
+cst_utterance *default_pos_tagger(cst_utterance *u)
+{
+    cst_item *word;
+    const cst_val *p;
+    const cst_cart *tagger;
+
+    p = get_param_val(u->features,"pos_tagger_cart",NULL);
+    if (p == NULL)
+        return u;
+    tagger = val_cart(p);
+
+    for (word=relation_head(utt_relation(u,"Word")); 
+	 word; word=item_next(word))
+    {
+        p = cart_interpret(word,tagger);
+        item_set_string(word,"pos",val_string(p));
+    }
+
+    return u;
+}
+
 cst_utterance *default_lexical_insertion(cst_utterance *u)
 {
     cst_item *word;
     cst_relation *sylstructure,*seg,*syl;
-    cst_lexicon *lex, *ulex = NULL;
-    const cst_val *p;
+    cst_lexicon *lex;
+    const cst_val *lex_addenda = NULL;
+    const cst_val *p, *wp = NULL;
     char *phone_name;
     char *stress = "0";
+    const char *pos;
     cst_val *phones;
     cst_item *ssword, *sssyl, *segitem, *sylitem, *seg_in_syl;
-    
+
     lex = val_lexicon(feat_val(u->features,"lexicon"));
-    if (feat_present(u->features, "user_lexicon"))
-	ulex = val_lexicon(feat_val(u->features, "user_lexicon"));
+    if (lex->lex_addenda)
+	lex_addenda = lex->lex_addenda;
 
     syl = utt_relation_create(u,"Syllable");
     sylstructure = utt_relation_create(u,"SylStructure");
     seg = utt_relation_create(u,"Segment");
-    
+
     for (word=relation_head(utt_relation(u,"Word")); 
 	 word; word=item_next(word))
     {
 	ssword = relation_append(sylstructure,word);
+        pos = ffeature_string(word,"pos");
 	phones = NULL;
+        wp = NULL;
+        
+        /*        printf("awb_debug word %s pos %s gpos %s\n",
+               item_feat_string(word,"name"),
+               pos,
+               ffeature_string(word,"gpos")); */
 
 	/* FIXME: need to make sure that textanalysis won't split
            tokens with explicit pronunciation (or that it will
@@ -408,10 +461,11 @@ cst_utterance *default_lexical_insertion(cst_utterance *u)
 	    phones = (cst_val *) item_feat(item_parent(item_as(word, "Token")), "phones");
 	else
 	{
-	    if (ulex)
-		phones = lex_lookup(ulex,item_feat_string(word, "name"),0);
-	    if (phones == NULL)
-		phones = lex_lookup(lex,item_feat_string(word,"name"),0);
+            wp = val_assoc_string(item_feat_string(word, "name"),lex_addenda);
+            if (wp)
+                phones = (cst_val *)val_cdr(val_cdr(wp));
+            else
+		phones = lex_lookup(lex,item_feat_string(word,"name"),pos);
 	}
 
 	for (sssyl=NULL,sylitem=NULL,p=phones; p; p=val_cdr(p))
@@ -424,23 +478,34 @@ cst_utterance *default_lexical_insertion(cst_utterance *u)
 	    }
 	    segitem = relation_append(seg,NULL);
 	    phone_name = cst_strdup(val_string(val_car(p)));
-	    if (phone_name[strlen(phone_name)-1] == '1')
+	    if (phone_name[cst_strlen(phone_name)-1] == '1')
 	    {
 		stress = "1";
-		phone_name[strlen(phone_name)-1] = '\0';
+		phone_name[cst_strlen(phone_name)-1] = '\0';
+	    }
+	    else if (phone_name[cst_strlen(phone_name)-1] == '0')
+	    {
+		stress = "0";
+		phone_name[cst_strlen(phone_name)-1] = '\0';
 	    }
 	    item_set_string(segitem,"name",phone_name);
 	    seg_in_syl = item_add_daughter(sssyl,segitem);
+#if 0
+            printf("awb_debug ph %s\n",phone_name);
+#endif
 	    if ((lex->syl_boundary)(seg_in_syl,val_cdr(p)))
 	    {
+#if 0
+                printf("awb_debug SYL\n");
+#endif
 		sylitem = NULL;
 		if (sssyl)
 		    item_set_string(sssyl,"stress",stress);
 	    }
 	    cst_free(phone_name);
 	}
-	/* FIXME: urgh argh ugly hack */
-	if (!item_feat_present(item_parent(item_as(word, "Token")), "phones"))
+	if (!item_feat_present(item_parent(item_as(word, "Token")), "phones")
+            && ! wp)
 	    delete_val(phones);
     }
 
@@ -477,21 +542,95 @@ cst_utterance *flat_prosody(cst_utterance *u)
 static cst_utterance *tokentosegs(cst_utterance *u)
 {
     cst_item *t;
-    cst_relation *seg;
+    cst_relation *seg, *syl, *sylstructure, *word;
+    cst_item *sylitem, *sylstructureitem, *worditem, *sssyl;
     cst_phoneset *ps;
 
     ps = val_phoneset(utt_feat_val(u, "phoneset"));
     /* Just copy tokens into the Segment relation */
     seg = utt_relation_create(u, "Segment");
-    for (t = relation_head(utt_relation(u, "Token")); t; t = item_next(t)) {
-	    cst_item *segitem = relation_append(seg, NULL);
-	    char const *pname = item_feat_string(t, "name");
+    syl = utt_relation_create(u, "Syllable");
+    word = utt_relation_create(u, "Word");
+    sylstructure = utt_relation_create(u, "SylStructure");
+    sssyl = sylitem = worditem = sylstructureitem = 0;
+    for (t = relation_head(utt_relation(u, "Token")); t; t = item_next(t)) 
+    {
+	cst_item *segitem = relation_append(seg, NULL);
+	char const *pname = item_feat_string(t, "name");
+	char *name = cst_strdup(pname);
 
-	    if (phone_id(ps, pname) == -1) {
-		    cst_errmsg("Phone `%s' not in phoneset\n", pname);
-		    cst_error();
-	    }
-	    item_set_string(segitem, "name", pname);
+	if (worditem == 0)
+	{
+	    worditem = relation_append(word,NULL);
+	    item_set_string(worditem, "name", "phonestring");
+	    sylstructureitem = relation_append(sylstructure,worditem);
+	}
+	if (sylitem == 0)
+	{
+	    sylitem = relation_append(syl,NULL);
+	    sssyl = item_add_daughter(sylstructureitem,sylitem);
+	}
+	
+	if (name[cst_strlen(name)-1] == '1')
+	{
+	    item_set_string(sssyl,"stress","1");
+	    name[cst_strlen(name)-1] = '\0';
+	}
+	else if (name[cst_strlen(name)-1] == '0')
+	{
+	    item_set_string(sssyl,"stress","0");
+	    name[cst_strlen(name)-1] = '\0';
+	}
+
+	if (cst_streq(name,"-"))
+	{
+	    sylitem = 0;  /* syllable break */
+	}
+	else if (phone_id(ps, name) == -1) 
+	{
+	    cst_errmsg("Phone `%s' not in phoneset\n", pname);
+	    cst_error();
+	}
+	else
+	{
+	    item_add_daughter(sssyl,segitem);
+	    item_set_string(segitem, "name", name);
+	}
+
+	cst_free(name);
     }
+
     return u;
 }
+
+int default_utt_break(cst_tokenstream *ts,
+		      const char *token,
+		      cst_relation *tokens)
+{
+    /* This is the default utt break functions, languages may override this */
+    /* This will be ok for some latin based languages */
+    const char *postpunct = item_feat_string(relation_tail(tokens), "punc");
+    const char *ltoken = item_name(relation_tail(tokens));
+
+    if (cst_strchr(ts->whitespace,'\n') != cst_strrchr(ts->whitespace,'\n'))
+	 /* contains two new lines */
+	 return TRUE;
+    else if (strchr(postpunct,':') ||
+	     strchr(postpunct,'?') ||
+	     strchr(postpunct,'!'))
+	return TRUE;
+    else if (strchr(postpunct,'.') &&
+	     (cst_strlen(ts->whitespace) > 1) &&
+	     strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZ",token[0]))
+	return TRUE;
+    else if (strchr(postpunct,'.') &&
+	     /* next word starts with a capital */
+	     strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZ",token[0]) &&
+	     /* last word isn't an abbreviation */
+	     !(strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZ",ltoken[cst_strlen(ltoken)-1])||
+	       ((cst_strlen(ltoken) < 4) &&
+		strchr("ABCDEFGHIJKLMNOPQRSTUVWXYZ",ltoken[0]))))
+	return TRUE;
+    else
+	return FALSE;
+}
